function [x, y, out] = pdNCG_TV(opts)
%PDNCG pdNCG solver for the TV-regularized problem
%
% min tau*||x||_TV + 0.5||Ax-b||_2^2.

%% Initialize output.
out.itr  = 1; % pdNCG iterations.
out.f = sparse(opts.maxpdNCGitr,1); % Objective values at every iteration.
out.t = sparse(opts.maxpdNCGitr,1); % time passed at every iteration.

%% Initialize data.
stagnation = 0;

% PCG data structure that carries everything the call to the
% preconditioned CG method needs.
PCG = [];

PCG.I = sparse(1:opts.n,1:opts.n,ones(opts.n,1));

%% Check if A'*A = I, i.e. denoising.
tmp = rand(opts.n1*opts.n2,1);
if norm(opts.A'*(opts.A*tmp)-tmp,1)/norm(tmp,1) < 1e-15;
    opts.reg = 1.0e-16;
    PCG.rho = 1;
else
    PCG.rho = 0.5;
end
clear tmp;

nwtn = zeros(opts.n,1);

x   = opts.x0;
Wt_re_x = opts.Wt_re*x;
Wt_im_x = opts.Wt_im*x;
Wtx = Wt_re_x +1i*Wt_im_x;

D        = 1./sqrt(opts.mu^2 + conj(Wtx).*Wtx);
DWt_re_x = D.*Wt_re_x;
DWt_im_x = D.*Wt_im_x;

% Initialize dual variables
y_re    = real(opts.y0);
y_im    = imag(opts.y0);

% Calculate the approximate Hessian of pseudo-Huber function.
scale_1 = D.*(1 - DWt_re_x.*y_re);
scale_2 = D.*(DWt_im_x.*y_re);
scale_3 = D.*(DWt_re_x.*y_im);
scale_4 = D.*(1 - DWt_im_x.*y_im);
H       = @(q) hesvec_TV(q,opts.Wt_re,opts.Wt_im,scale_1,scale_2,...
                         scale_3,scale_4);
                     
% Give some data to PCG structure.
PCG.scale_1 = scale_1;
PCG.scale_2 = scale_2;
PCG.scale_3 = scale_3;
PCG.scale_4 = scale_4;

% Calculate gradient at x0.
g_ls = opts.A.'*(opts.A*x-opts.b);
g_ph = opts.Wt_re'*DWt_re_x + opts.Wt_im'*DWt_im_x;
g    = opts.tau.*g_ph + g_ls;

% Fidelity and regularization functions.
f_ls    = @(q) 0.5*norm(opts.A*q-opts.b)^2;
f_huber = @(q) opts.tau*sum(sqrt(opts.mu^2 + abs(opts.Wt_re*q + ...
                                           1i*opts.Wt_im*q).^2) - opts.mu);
     
%% Print
if opts.verbose >= 1
    fprintf('%5s | %5s    | %5s | %5s |  %5s  \n',...
        'pdNCG itr', 'a', '|d|_x/|x|','PCG itr', 'PCG rs');
end

%% Start Newton conjugate gradients.
while 1        
    % Define regularized second-order derivative.
    Hessian = @(q) opts.tau.*H(q) + opts.A'*(opts.A*q) + opts.reg*q;   
    
    % Calculate Newton direction. 
    [nwtn, rlrs, pcgi, preconditioner] = PCG_wrapper(PCG, Hessian, g, opts);   
    
    % Call user function.
    if isfield(opts, {'user_fcn'}) 
        opts.user_fcn(opts,x,y_re + 1i*y_im,@(q) f_huber(q) + f_ls(q),g,Hessian,preconditioner,nwtn);
    end   
    
    normN_scaled = -(nwtn'*g + opts.reg*norm(nwtn)^2)/norm(x);    
        
    Wt_re_nwtn = opts.Wt_re*nwtn;
    Wt_im_nwtn = opts.Wt_im*nwtn;
    dy_re = scale_1.*Wt_re_nwtn  + scale_2.*Wt_im_nwtn -(y_re - DWt_re_x);
    dy_im = scale_4.*Wt_im_nwtn  + scale_3.*Wt_re_nwtn -(y_im - DWt_im_x);
        
    y_tilde_re = y_re + dy_re;
    y_tilde_im = y_im + dy_im;
    y_tilde    = y_tilde_re + 1i*y_tilde_im;
    y_bar      = min(1./abs(y_tilde),1).*y_tilde;    
    y_re = real(y_bar);
    y_im = imag(y_bar);
    
    % Calculate stepsize and update variables.
    [x, a, f_norm1_new] = simple_bt(x, nwtn, f_ls, f_huber, normN_scaled, opts);
    
    % Store objective function and corresponding time.
    out.f(out.itr) = f_norm1_new;
    out.t(out.itr) = toc; 
 
    % The PCG methods needs to know the stepsize.
    PCG.a = a;
    
    % Printing info.
    if opts.verbose >= 1
        fprintf('  %4d    | %1.2e |  %8s | %4d    | %1.2e \n',...
            out.itr, a, num2str(normN_scaled,'%1.2e\n'), pcgi, rlrs);
    end
        
    % Calculate the approximate Hessian of pseudo-Huber function.
    Wt_re_x = opts.Wt_re*x;
    Wt_im_x = opts.Wt_im*x;
    Wtx = Wt_re_x +1i*Wt_im_x;

    D     = 1./sqrt(opts.mu^2 + conj(Wtx).*Wtx);
    DWt_re_x  = D.*Wt_re_x;
    DWt_im_x  = D.*Wt_im_x;
    scale_1 = D.*(1 - DWt_re_x.*y_re);
    scale_2 = D.*(DWt_im_x.*y_re);
    scale_3 = D.*(DWt_re_x.*y_im);
    scale_4 = D.*(1 - DWt_im_x.*y_im);
    H       = @(q) hesvec_TV(q,opts.Wt_re,opts.Wt_im,scale_1,scale_2,...
                             scale_3,scale_4);
                         
    % Give some data to PCG structure.
    PCG.scale_1 = scale_1;
    PCG.scale_2 = scale_2;
    PCG.scale_3 = scale_3;
    PCG.scale_4 = scale_4;

    % Calculate gradient at x.
    g_ls  = opts.A.'*(opts.A*x-opts.b);
    g_ph  = opts.Wt_re'*DWt_re_x + opts.Wt_im'*DWt_im_x;
    g     = opts.tau.*g_ph + g_ls;
    
    if isfield(opts, {'user_stop'}) && opts.last_cont_step
        if opts.user_stop(opts,x,y_re + 1i*y_im,...
                                  @(q) f_huber(q) + f_ls(q),g,Hessian,nwtn)
            out.message = 'Converged';
            break;
        end
    else
        if normN_scaled <= opts.tol
            out.message = 'Converged';
            break; 
        end         
    end
    
    % Maximum iterations reached.
    if out.itr > opts.maxpdNCGitr
        stagnation = 1;
    end
    
    if stagnation == 1
       out.message = 'Maximum iterations of pdNCG reached';
       if opts.verbose >= 1
          fprintf('Maximum iterations of pdNCG reached, termination.\n'); 
          fprintf('Possible fixes: \n');
          fprintf('a) increase opts.maxpdNCGitr. Performs more pdNCG iterations.\n');
          fprintf('b) decrease opts.tolpcg. Calculates better directions. \n');
          fprintf('c) increase opts.mxitrpcg. \n');
       end
       break;
    end 
    
    % Update iteration counter.
    out.itr = out.itr + 1;
end
%% End of primal-dual Newton conjugate gradients.

out.f = out.f(1:out.itr);
out.t = out.t(1:out.itr);

% Return final complex dual variable.
y = y_re + 1i*y_im;

% Call user function.
if isfield(opts, {'user_fcn'}) 
    opts.user_fcn(opts,x,y_re + 1i*y_im,@(q) f_huber(q) + f_ls(q),g,Hessian,preconditioner,nwtn);
end  