% This script reproduces Table 7.2 in [1]
%
% [1] I. Dassios, K. Fountoulakis and J. Gondzio, "A Preconditioner for a
%     Primal-Dual Newton Conjugate Gradients Method for Compressed Sensing 
%     Problems". Technical Report ERGO 14-021.
%
% ** Signal Processing Toolbox is required **
%
% This file should take about 4.146310e+03 sec. CPU time to be completed.

clear;
tt_file = cputime;

problems = [1,2,3,4,5];

taus = [1,5,10,15,30,45];
mus_tval = [1,5.0e-1,8.9e-2,8.9e-2,8.05e-2,8.05e-2];

nsize = [64; 128; 256; 512; 1024; 2048]';

for prob=problems,
    
    fprintf('Problem size: %d \n \n',nsize(prob));

    randn('state',245);
    rand('state',245);

    n1 = nsize(prob); 
    n2 = n1;
    N = n1*n2;
    n = max(n1,n2);

    % Select phantom image
    x = phantom(n);

    x = x(1:n1,1:n2);
    x_original = x;
    global mat;
    mat = @(x) reshape(x,n1,n2);
    vec = @(x) x(:);

    % Partial Fourier measurements.
    M = round( N / 4 );
    omega = randn(N,1);
    [temp,omega] = sort(omega);
    omega = sort( omega(1:M) );
    downsample = @(x) x(omega);
    SS.type = '()'; SS.subs{1} = omega; SS.subs{2} = ':';
    upsample = @(x) subsasgn( zeros(N,size(x,2)),SS,x);

    % take partial 2D DCT measurements.  We have Af(At) = identity (not
    % vice-versa of course). Randomly permuted.
    rp = randperm(N);
    [~,rp_inv] = sort(rp);
    rpF = @(x) x(rp);
    rp_invF = @(x) x(rp_inv);
        
    my_dct2 = @(x) dct(dct(x).').';
    my_idct2= @(x) idct(idct(x).').';
    Af = @(x) downsample(vec( my_dct2( mat(rpF(mat(x) )) ) ) );
    At = @(y) vec( rp_invF( my_idct2( mat( upsample( y ) ) ) ) );

    % treat X as n1*n2 x 1 vector?
    VECTOR = true;
    % or treat X as n1 x n2 matrix?
    % VECTOR = false;

    if VECTOR
        A = linop_handles([M,N], Af, At );
        VEC = vec;
    else
        Af = @(x) vec(Af(x));
        At = @(x) mat(At(x));
        A = linop_handles({[n1,n2],[M,1]}, Af, At );
        VEC = @(x) x;
    end
    
    myAwgn = @(x,snr) x + ...
        10^( (10*log10(sum(abs(x(:)).^2)/length(x(:))) - snr)/20 )*randn(size(x));
    
    b_original = Af(vec(x_original));
    snr = 10;  % SNR in dB
    x_noisy = myAwgn( x_original, snr );
    b   = Af( vec(x_noisy) );
    EPS = 0.9*norm(b-b_original);

    mu  = 1.0e-5*norm( linop_TV(x_original) ,Inf);

    x_ref   = x_original;

    % %% Print Phantom clean.
    % figure(1)
    % imshow( x_original );
    % maxI    = 1; % max pixel value
    % PSNR    = @(x) 20*log10(maxI*sqrt(N)/norm(vec(x)-vec(x_original) ) );
    % title(sprintf('Phantom noiseless, PSNR is %.1f dB',...
    %     PSNR( x_original )));
    % 
    % %% Print Phantom noisy.
    % figure(2)
    % imshow( x_noisy );
    % maxI    = 1; % max pixel value
    % PSNR    = @(x) 20*log10(maxI*sqrt(N)/norm(vec(x)-vec(x_original) ) );
    % title(sprintf('Phantom noisy, PSNR is %.1f dB',...
    %     PSNR( x_noisy )));

    sz              = A([],0);
    M               = sz{2}(1);
    N               = sz{1}(1);
    [n1,n2]         = size(x_original);
    norm_x_ref      = norm(x_ref,'fro');
    er_ref          = @(x) norm(vec(x)-vec(x_ref))/norm_x_ref;

    % fprintf('\tA is %s, %d x %d, signal has SNR %d dB\n', 'partial Fourier', M, N, round(snr) );

    %% Call TFOCS solver.
    er              = er_ref;  
    opts = [];
    opts.restart    = 1000;
    opts.errFcn     = @(f,dual,primal) er(primal);
    opts.maxIts     = 250;

    x0 = At(b);
    z0  = [];   % we don't have a good guess for the dual
    if VECTOR
        W   = linop_TV( [n1,n2] );
    else
        W   = linop_TV( {n1,n2} );
    end
    normW           = linop_TV( [n1,n2], [], 'norm' );
    opts.normW2     = normW^2;
    
    opts.printEvery = 0;
    contOpts.verbose = false;
    
    %% Call constrained TFOCS, also gives the an approximation to optimal tau
    tstart = cputime;
    solver = @(mu,x0,z0,opts) solver_sBPDN_W( A, W, b, EPS, mu, x0, z0, opts);
    [ x_con, out, ~ ] = continuation(solver,5*mu,x0(:),z0,opts,contOpts);
    time_TFOCS_con = cputime - tstart;
        
    fprintf('TFOCS constrained, Solution has %d nonzeros.  Error vs. original solution is %.2e\n',...
            nnz(x_con), er(x_con) );
    x_con = mat(x_con);
    %maxI    = 1; % max pixel value
    %PSNR    = @(x) 20*log10(maxI*sqrt(N)/norm(vec(x)-vec(x_original) ) );
    [psnr_val, snr] = psnr(x_con, x_original);
    PSNR_x_con = fix(psnr_val*10)/10;
    fprintf('TFOCS constrained, PSNR is %.1f dB, CPU time: %.1f sec. \n \n',...
            PSNR_x_con, time_TFOCS_con);

    tstart = cputime;
    % [ x, out, optsOut ] = solver_sBPDN_W( A, W, b, EPS, mu, VEC(x0), z0, opts
    % );
    tau = taus(prob)/norm(out.dual{1});
    solver = @(mu,x0,z0,opts) my_solver_sBPDN_W( A, W, b, tau, mu*tau, x0, z0, opts);
    [ x, ~, optsOut ] = continuation(solver,5*mu,x0(:),z0,opts,contOpts);
    time_TFOCS = cputime - tstart;
    %fprintf('\n \n');

    %% Print TFOCS
    fprintf('TFOCS, Solution has %d nonzeros.  Error vs. original solution is %.2e\n',...
        nnz(x), er(x) );
    x = mat(x);
    %maxI    = 1; % max pixel value
    %PSNR    = @(x) 20*log10(maxI*sqrt(N)/norm(vec(x)-vec(x_original) ) );
    [psnr_val, snr] = psnr(x, x_original);
    PSNR_x = fix(psnr_val*10)/10;
    fprintf('TFOCS, PSNR is %.1f dB, CPU time: %.1f sec. \n \n',...
        PSNR_x, time_TFOCS);

    %% Call TVAL3.
    optstval.mu = 1/tau;
    optstval.beta = 2^8;
    optstval.beta0 = 1;
    optsval.mu0 = 1;
    optstval.tol = 1E-3;
    optstval.maxit = 3000;
    optstval.maxcnt = 3000;
%     optstval.TVnorm = 1;
    optstval.TVL2 = true;
    optstval.isreal = true;
%     optstval.init = mat(At(b));
    optstval.nonneg = false;
    optstval.disp = false;
    optstval.PSNR_tfocs = PSNR_x;
    optstval.x_original = x_original;
    
    tstart = cputime;
    [x_tval, out_tval] = TVAL3(A,b,n,n,optstval);
    time_TVAL = cputime - tstart;

    %% Print TVAL
    fprintf('TVAL3, Solution has %d nonzeros.  Error vs. original solution is %.2e\n',...
        nnz(x_tval), er(x_tval) );
    x_TVAL = mat(x_tval);
    %maxI    = 1; % max pixel value
    %PSNR    = @(x) 20*log10(maxI*sqrt(N)/norm(vec(x)-vec(x_original) ) );
    [psnr_val, snr] = psnr(x_TVAL, x_original);
    PSNR_x_TVAL = fix(psnr_val*10)/10;
    fprintf('TVAL3, PSNR is %.1f dB, CPU time: %.1f sec. \n \n',...
        PSNR_x_TVAL, time_TVAL);

    %% Call pdNCG solver.
    ApdNCG  = A_operator(@(w) A(w,1), @(w) A(w,2));

    optspdNCG.n1  = n;
    optspdNCG.n2  = n;
    optspdNCG.b   = b;
    optspdNCG.A   = ApdNCG;
    optspdNCG.tau = tau;
    optspdNCG.tol = 1.0e-1;
    optspdNCG.mu = 1.0e-5;
    optspdNCG.maxpdNCGitr = 5;
    optspdNCG.problem  = 'tv';
    optspdNCG.verbose  = 0;
    optspdNCG.user_stop = @(opts,primal,dual,f,g,H,d) PSNR_x <= psnr(mat(primal), x_original);

    tstart = cputime;
    [xDC, dual, out_pdNCG] = pdNCGinterface(optspdNCG);
    time_pdNCG = cputime - tstart;

    %% Print pdNCG
    fprintf('pdNCG, solution has %d nonzeros.  Error vs. original solution is %.2e\n',...
        nnz(xDC), er(xDC) );
    xDC = mat(real(xDC));
    %maxI    = 1; % max pixel value
    %PSNR    = @(x) 20*log10(maxI*sqrt(N)/norm(vec(x)-vec(x_original) ) );
    [psnr_val, snr] = psnr(xDC, x_original);
    PSNR_xDC = fix(psnr_val*10)/10;
    fprintf('pdNCG, PSNR is %.1f dB, CPU time is %.1f \n \n',...
        PSNR_xDC, time_pdNCG);
    
    %% TwIST
    tv_iters = 10;
    Psi = @(x,th)  tvdenoise(x,2/th,tv_iters);
    
    % set the penalty function, to compute the objective
    Phi = @(x) TVnorm(x);

    tolA = PSNR_x;
    
    tstart = cputime;
    % -- TwIST ---------------------------
    % stop criterium:  the relative change in the objective function 
    % falls below 'ToleranceA'         
    [x_twist,dummy,obj_twist,times_twist,dummy,mse_twist]= ...
         TwIST(b,Af,tau,x_original,...
         'Lambda', 0.04, ...
         'AT', At, ...
         'Psi', Psi, ...
         'Phi',Phi, ...
         'Monotone',1,...
         'MaxiterA', 3000, ...
         'Initialization',0,...
         'StopCriterion',4,...
       	 'ToleranceA',tolA,...
         'Verbose', 0);
     time_TwIST = cputime - tstart;
        
     %% Print TwIST
     fprintf('TwIST, solution has %d nonzeros.  Error vs. original solution is %.2e\n',...
     nnz(x_twist), er(x_twist) );
     x_twist = mat(real(x_twist));
     %maxI    = 1; % max pixel value
     %PSNR    = @(x) 20*log10(maxI*sqrt(N)/norm(vec(x)-vec(x_original) ) );
     [psnr_val, snr] = psnr(x_twist, x_original);
     PSNR_TwIST = fix(psnr_val*10)/10;
     fprintf('TwIST, PSNR is %.1f dB, CPU time is %.1f \n \n',...
     PSNR_TwIST, time_TwIST);     
     
end

fprintf('Elapsed CPU time for this file: %d sec. \n', cputime - tt_file);
clear all;

